"""Module which extends the scenepic Canvas3D type."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from typing import Union, Mapping

from ._scenepic import (
    Canvas2D,
    LayerSettings,
)


def set_layer_settings(self, layer_settings: Mapping[str, Union[dict, LayerSettings]]):
    """Specify the visibilities and opacities of certain mesh layers.

    Description:
        Each Mesh object can optionally be part of a user-identified layer
        (see Mesh constructor).
        Calling set_layer_settings will result in an additional drop-down
        selector in the ScenePic user interface.

    Args:
        self (Canvas2D): the self reference
        layer_settings (Mapping[str, Union[dict, LayerSettings]]): a mapping from
            layer IDs to layer settings.
    """
    updated_settings = {}
    for key, value in layer_settings.items():
        if isinstance(value, dict):
            if "visible" in value:
                print("`visible` keyword has been deprecated")
            else:
                updated_settings[key] = LayerSettings(**value)

    self.set_layer_settings_(updated_settings)


Canvas2D.set_layer_settings = set_layer_settings



"""Set of unit tests formed from the Python Tutorial example."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import json
import math

import numpy as np
import scenepic as sp


def test_scene_and_canvas_basics():
    """Test the most basic functionality."""
    scene = sp.Scene()
    scene.create_canvas_3d(width=300, height=300)
    scene.create_canvas_3d(width=100, height=300)


def test_meshes_and_frames():
    """Test creating meshes and frames."""
    scene = sp.Scene()

    my_first_mesh = scene.create_mesh(shared_color=sp.Color(1.0, 0.0, 1.0))
    my_first_mesh.add_cube(transform=sp.Transforms.Scale(0.1))
    my_first_mesh.add_cube(transform=np.dot(sp.Transforms.Translate(
        [-1.0, 1.0, -1.0]), sp.Transforms.Scale(0.5)))
    my_first_mesh.add_sphere(
        transform=sp.Transforms.Translate([1.0, 1.0, 1.0]))

    canvas = scene.create_canvas_3d(width=300, height=300)

    for i in range(10):
        frame = canvas.create_frame()
        frame.add_mesh(my_first_mesh, transform=sp.Transforms.Translate(
            [i / 10.0, 0.0, 0.0]))
        mesh2 = scene.create_mesh(shared_color=sp.Color(
            1.0, 0.0, 0.0), camera_space=True)
        mesh2.add_cube(transform=np.dot(sp.Transforms.Translate(
            [0.0, 0.0, -5.0]), sp.Transforms.Scale(0.5)))
        frame.add_mesh(mesh2)
        label = scene.create_label(text="Hi", color=sp.Colors.White,
                                   size_in_pixels=80, offset_distance=0.6, camera_space=True)
        frame.add_label(label=label, position=[0.0, 0.0, -5.0])


def test_point_clouds_1():
    """Test creating point clouds."""
    scene = sp.Scene()

    mesh = scene.create_mesh(shared_color=sp.Color(0, 1, 0))
    mesh.add_cube()
    mesh.apply_transform(sp.Transforms.Scale(0.01))
    mesh.enable_instancing(positions=2 * np.random.rand(10000, 3) - 1)

    canvas = scene.create_canvas_3d(
        width=300, height=300, shading=sp.Shading(bg_color=sp.Colors.White))
    frame = canvas.create_frame()
    frame.add_mesh(mesh)


def test_point_clouds_2():
    """More testing of point clouds."""
    disc_thickness = 0.2
    normal_length = 1.5
    point_size = 0.1

    sphere_mesh = sp.Mesh()
    sphere_mesh.add_sphere(transform=sp.Transforms.Scale(2.0), color=sp.Color(1.0, 0.0, 0.0))
    num_vertices = sphere_mesh.count_vertices()
    points = sphere_mesh.vertex_buffer["pos"]
    normals = sphere_mesh.vertex_buffer["norm"]

    rotations = np.zeros((num_vertices, 4))
    for i in range(0, num_vertices):
        rotations[i, :] = sp.Transforms.QuaternionToRotateXAxisToAlignWithAxis(
            normals[i, :])

    colors = np.random.rand(num_vertices, 3)

    scene = sp.Scene()

    mesh = scene.create_mesh(shared_color=sp.Color(0, 1, 0), double_sided=True)

    mesh.add_disc(segment_count=20, transform=sp.Transforms.Scale(
        [disc_thickness, 1.0, 1.0]))
    mesh.add_thickline(start_point=np.array([disc_thickness * 0.5, 0.0, 0.0]),
                       end_point=np.array([normal_length, 0.0, 0.0]),
                       start_thickness=0.2,
                       end_thickness=0.1)
    mesh.apply_transform(sp.Transforms.Scale(point_size))

    mesh.enable_instancing(positions=points, rotations=rotations, colors=colors)

    canvas = scene.create_canvas_3d(width=300, height=300)
    frame = canvas.create_frame()
    frame.add_mesh(mesh)


def test_misc_meshes():
    """Test various miscellaneous meshes."""
    scene = sp.Scene()

    mesh1 = scene.create_mesh(layer_id="Sphere+")
    mesh1.add_cube(color=sp.Color(1.0, 0.0, 0.0),
                   transform=sp.Transforms.Translate([-2.0, 0.0, -2.0]))
    mesh1.add_uv_sphere(color=sp.Color(0.0, 0.0, 1.0),
                        transform=np.dot(sp.Transforms.Translate([-1.0, 1.0, 0.0]),
                                         sp.Transforms.Scale(1.8)),
                        fill_triangles=False,
                        add_wireframe=True)
    mesh1.add_icosphere(color=sp.Color(0.0, 1.0, 1.0),
                        transform=np.dot(sp.Transforms.Translate([2.0, 1.0, 0.0]),
                                         sp.Transforms.Scale(1.8)),
                        fill_triangles=False,
                        add_wireframe=True,
                        steps=2)

    mesh2 = scene.create_mesh(layer_id="Coords")
    mesh2.add_coordinate_axes(transform=sp.Transforms.Translate([0.0, 0.0, 0.0]))

    cube_verts = np.array([[-0.5, -0.5, -0.5],
                           [+0.5, -0.5, -0.5],
                           [-0.5, +0.5, -0.5],
                           [+0.5, +0.5, -0.5],
                           [-0.5, -0.5, +0.5],
                           [+0.5, -0.5, +0.5],
                           [-0.5, +0.5, +0.5],
                           [+0.5, +0.5, +0.5]])
    cube_tris = np.array([[0, 2, 3],
                          [0, 3, 1],
                          [1, 3, 7],
                          [1, 7, 5],
                          [4, 5, 7],
                          [4, 7, 6],
                          [4, 6, 2],
                          [4, 2, 0],
                          [2, 6, 7],
                          [2, 7, 3],
                          [4, 0, 1],
                          [4, 1, 5]])
    cube_verts_a, cube_tris_a = sp.LoopSubdivStencil(cube_tris, 2, False).apply(cube_verts)
    cube_verts_b, cube_tris_b = sp.LoopSubdivStencil(cube_tris, 2, True).apply(cube_verts)
    mesh3 = scene.create_mesh(shared_color=sp.Color(1.0, 0.8, 0.8))
    mesh3.add_mesh_without_normals(cube_verts, cube_tris,
                                   transform=sp.Transforms.Translate([-1.0, 0.0, 0.0]))
    mesh3.add_mesh_without_normals(cube_verts_a, cube_tris_a)
    mesh3.add_mesh_without_normals(cube_verts_b, cube_tris_b,
                                   transform=sp.Transforms.Translate([+1.0, 0.0, 0.0]))

    mesh4 = scene.create_mesh()
    num_segs = 7000
    positions = np.cumsum(np.random.rand(num_segs, 3) * 0.2, axis=0)
    colored_points = np.concatenate(
        (positions, np.random.rand(num_segs, 3)), axis=1)
    mesh4.add_lines(colored_points[0:-1, :], colored_points[1:, :])
    mesh4.add_camera_frustum(color=sp.Color(1.0, 1.0, 0.0))

    canvas1 = scene.create_canvas_3d(width=300, height=300)
    canvas2 = scene.create_canvas_3d(width=300, height=300)

    scene.link_canvas_events(canvas1, canvas2)

    canvas1.set_layer_settings({
        "Coords": {"visible": False},
        "Sphere+": {"visible": True}
    })

    canvas1.create_frame(meshes=[mesh1, mesh2])
    canvas2.create_frame(meshes=[mesh2, mesh3])
    canvas2.create_frame(meshes=[mesh4, mesh1])


def test_images_and_textures(asset):
    """Test loading images."""
    scene = sp.Scene()

    image1 = scene.create_image(image_id="PolarBear")
    image1.load(asset("PolarBear.png"))

    texture = scene.create_image(image_id="texture")
    texture.load(asset("uv.png"))

    cam_space_mesh = scene.create_mesh(shared_color=sp.Color(1.0, 0.0, 0.0), camera_space=True)
    cam_space_mesh.add_sphere(transform=np.dot(sp.Transforms.Translate([10, -10, -20.0]),
                                               sp.Transforms.Scale(1.0)))

    sphere = scene.create_mesh(texture_id=texture.image_id, nn_texture=False)
    sphere.add_icosphere(steps=4)

    cube = scene.create_mesh(texture_id=texture.image_id)
    transform = sp.Transforms.translate([-1, 0, 0]) @ sp.Transforms.scale(0.5)
    cube.add_cube(transform=transform)

    canvas = scene.create_canvas_3d(shading=sp.Shading(bg_color=sp.Colors.White))
    mesh1 = scene.create_mesh(texture_id="PolarBear")
    mesh1.add_image()

    n_frames = 20
    for i in range(n_frames):
        angle = 2 * math.pi * i / n_frames
        cos, sin = math.cos(angle), math.sin(angle)

        axis = np.array([1.0, 0.0, 1.0])
        axis /= np.linalg.norm(axis)
        focus_point = sp.FocusPoint([cos, sin, 0], orientation_axis_angle=axis * angle)

        mesh = scene.create_mesh()
        mesh.add_coordinate_axes(transform=np.dot(
            sp.Transforms.Translate(focus_point.position),
            sp.Transforms.RotationMatrixFromAxisAngle(axis, angle)))

        im_size = 15
        im_data = np.random.rand(im_size, im_size, 4)
        im_data[:, :, 3] = 0.5 + 0.5 * im_data[:, :, 3]

        image_b = scene.create_image()
        image_b.from_numpy(im_data)
        mesh_b = scene.create_mesh(texture_id=image_b, is_billboard=True, use_texture_alpha=True)
        mesh_b.add_image(transform=np.dot(sp.Transforms.Scale(2.0),
                                          sp.Transforms.Translate([0, 0, -1])))

        frame = canvas.create_frame(focus_point=focus_point)
        frame.add_mesh(mesh1, transform=sp.Transforms.Translate([cos, sin, 0]))
        frame.add_mesh(mesh_b, transform=np.dot(sp.Transforms.Scale(i * 1.0 / n_frames),
                                                sp.Transforms.Translate([-cos, -sin, 0])))
        frame.add_mesh(cam_space_mesh)
        frame.add_mesh(sphere, transform=sp.Transforms.rotation_about_y(np.pi * 2 * i / n_frames))
        frame.add_mesh(cube, transform=sp.Transforms.rotation_about_y(-np.pi * 2 * i / n_frames))
        frame.add_mesh(mesh)


def test_2d_canvases(asset):
    """Tests for 2D canvases."""
    scene = sp.Scene()

    image1 = scene.create_image(image_id="PolarBear")
    image1.load(asset("PolarBear.png"))

    image2 = scene.create_image(image_id="Random")
    image2.from_numpy(np.random.rand(20, 30, 3) * 128 / 255.0)

    canvas1 = scene.create_canvas_2d(width=400, height=300, background_color=sp.Colors.White)
    canvas1.create_frame().add_image(image1, "fit")
    canvas1.create_frame().add_image(image1, "fill")
    canvas1.create_frame().add_image(image1, "stretch")
    canvas1.create_frame().add_image(image1, "manual", x=50, y=50, scale=0.3)

    canvas2 = scene.create_canvas_2d(width=300, height=300)
    frame = canvas2.create_frame()
    frame.add_image(image2, "fit")
    frame.add_image(image1, "manual", x=30, y=30, scale=0.2)
    frame.add_circle(200, 200, 40, fill_color=sp.Colors.Black,
                     line_width=10, line_color=sp.Colors.Blue)
    frame.add_rectangle(200, 100, 50, 25,
                        fill_color=sp.Colors.Green, line_width=0)
    frame.add_text("Hello World", 30, 100, sp.Colors.White, 100, "segoe ui light")


def test_opacity_and_labels():
    """Test opacity and label functionality."""
    np.random.seed(55)

    scene = sp.Scene()
    canvas = scene.create_canvas_3d(width=700, height=700)
    frame = canvas.create_frame()

    layer_settings = {"Labels": {"opacity": 1.0}}
    num_objects = 20
    for i in range(num_objects):
        geotype = np.random.randint(2)
        color = np.random.rand(3)
        size = 0.3 * np.random.rand() + 0.2
        position = 3.0 * np.random.rand(3) - 1.5
        opacity = 1.0 if np.random.randint(2) == 0 else np.random.uniform(0.45, 0.55)

        layer_id = "Layer" + str(i)
        mesh = scene.create_mesh(shared_color=color, layer_id=layer_id)
        layer_settings[layer_id] = {"opacity": opacity}
        if geotype == 0:
            mesh.add_cube()
        elif geotype == 1:
            mesh.add_sphere()
        mesh.apply_transform(sp.Transforms.Scale(size))
        mesh.apply_transform(sp.Transforms.Translate(position))
        frame.add_mesh(mesh)

        text = "{0:0.2f} {1:0.2f} {2:0.2f} {3:0.2f}".format(
            color[0], color[1], color[2], opacity)
        horizontal_align = ["left", "center", "right"][np.random.randint(3)]
        vertical_align = ["top", "middle", "bottom"][np.random.randint(3)]
        if geotype == 0:
            if horizontal_align != "center" and vertical_align != "middle":
                offset_distance = size * 0.7
            else:
                offset_distance = size * 0.9
        else:
            if horizontal_align != "center" and vertical_align != "middle":
                offset_distance = size * 0.5 * 0.8
            else:
                offset_distance = size * 0.6

        label = scene.create_label(text=text,
                                   color=color,
                                   layer_id="Labels",
                                   font_family="consolas",
                                   size_in_pixels=80 * size,
                                   offset_distance=offset_distance,
                                   vertical_align=vertical_align,
                                   horizontal_align=horizontal_align)
        frame.add_label(label=label, position=position)

    canvas.set_layer_settings(layer_settings)


def test_animation0(asset):
    scene = sp.Scene()
    canvas = scene.create_canvas_3d(width=700, height=700)

    jelly_mesh = sp.load_obj(asset("jelly.obj"))
    texture = scene.create_image("texture")
    texture.load(asset("jelly.png"))

    base_mesh = scene.create_mesh("jelly_base")
    base_mesh.texture_id = texture.image_id
    base_mesh.use_texture_alpha = True
    base_mesh.add_mesh(jelly_mesh)

    def random_linspace(min_val, max_val, num_samples):
        vals = np.linspace(min_val, max_val, num_samples)
        np.random.shuffle(vals)
        return vals

    marbles = scene.create_mesh("marbles_base")
    num_marbles = 10
    marbles.add_sphere(sp.Colors.White, transform=sp.Transforms.Scale(0.2))
    marble_positions = np.zeros((num_marbles, 3), np.float32)
    marble_positions[:, 0] = random_linspace(-0.6, 0.6, num_marbles)
    marble_positions[:, 2] = random_linspace(-1.0, 0.7, num_marbles)
    marble_offsets = np.random.uniform(0, 2 * np.pi, size=num_marbles).astype(np.float32)
    marble_colors_start = np.random.uniform(0, 1, size=(num_marbles, 3)).astype(np.float32)
    marble_colors_end = np.random.uniform(0, 1, size=(num_marbles, 3)).astype(np.float32)
    marbles.enable_instancing(marble_positions, colors=marble_colors_start)

    for i in range(60):
        positions = jelly_mesh.positions.copy()
        delta_x = (positions[:, 0] + 0.0838 * i) * 10
        delta_z = (positions[:, 2] + 0.0419 * i) * 10
        positions[:, 1] = positions[:, 1] + 0.1 * (np.cos(delta_x) + np.sin(delta_z))

        jelly_update = scene.update_mesh_positions("jelly_base", positions)
        frame = canvas.create_frame(meshes=[jelly_update])

        marble_y = np.sin(0.105 * i + marble_offsets)
        positions = np.stack([marble_positions[:, 0], marble_y, marble_positions[:, 2]], -1)
        alpha = ((np.sin(marble_y) + 1) * 0.5).reshape(-1, 1)
        beta = 1 - alpha
        colors = alpha * marble_colors_start + beta * marble_colors_end
        marbles_update = scene.update_instanced_mesh("marbles_base", positions, colors=colors)
        frame.add_mesh(marbles_update)

    scene.quantize_updates()


def test_animation1():
    scene = sp.Scene()

    butterflies = scene.create_mesh("butterflies", double_sided=True)
    butterflies.add_quad(sp.Colors.Blue, [0, 0, 0], [0.1, 0, 0.04], [0.08, 0, -0.06], [0.015, 0, -0.03])

    rotate_back = sp.Transforms.quaternion_from_axis_angle([1, 0, 0], -np.pi / 6)

    num_butterflies = 100
    num_anim_frames = 20

    start_frames = np.random.randint(0, num_anim_frames, num_butterflies)
    rot_angles = np.random.uniform(-1, 1, num_butterflies)

    rotations = np.zeros((num_butterflies * 2, 4), np.float32)
    positions = np.random.uniform(-1, 1, (num_butterflies * 2, 3))
    colors = np.random.random((num_butterflies * 2, 3))

    for b, angle in enumerate(rot_angles):
        rot = sp.Transforms.quaternion_from_axis_angle([0, 1, 0], angle)
        rotations[2 * b] = rotations[2 * b + 1] = rot

        dx = np.sin(angle) * 0.1
        dy = positions[2 * b + 1, 1] - positions[2 * b, 1]
        dy = np.sign(angle) * min(abs(angle), 0.1)
        dz = np.cos(angle) * 0.1
        positions[2 * b + 1] = positions[2 * b] + [dx, dy, dz]

    butterflies.enable_instancing(positions, rotations, colors)

    canvas = scene.create_canvas_3d("main", 700, 700)
    canvas.shading = sp.Shading(sp.Colors.White)

    start = -np.pi / 6
    end = np.pi / 2
    delta = (end - start) / (num_anim_frames // 2 - 1)

    animation = []
    for i in range(num_anim_frames):
        frame_positions = np.zeros_like(positions)
        frame_rotations = np.zeros_like(rotations)
        frame_colors = np.zeros_like(colors)

        for b, start_frame in enumerate(start_frames):
            frame = (i + start_frame) % num_anim_frames
            if frame < num_anim_frames // 2:
                angle = start + delta * frame
            else:
                angle = end + delta * (frame - num_anim_frames // 2)

            right = sp.Transforms.quaternion_from_axis_angle([0, 0, 1], angle)
            right = sp.Transforms.quaternion_multiply(rotate_back, right)
            right = sp.Transforms.quaternion_multiply(rotations[2 * b], right)
            left = sp.Transforms.quaternion_from_axis_angle([0, 0, 1], np.pi - angle)
            left = sp.Transforms.quaternion_multiply(rotate_back, left)
            left = sp.Transforms.quaternion_multiply(rotations[2 * b + 1], left)
            frame_rotations[2 * b] = right
            frame_rotations[2 * b + 1] = left

            progress = np.sin((frame * 2 * np.pi) / num_anim_frames)
            progress = (progress + 1) * 0.5

            pos = (1 - progress) * positions[2 * b] + progress * positions[2 * b + 1]
            pos[1] -= np.sin(angle) * 0.02
            frame_positions[2 * b : 2 * b + 2, :] = pos

            color = (1 - progress) * colors[2 * b] + progress * colors[2 * b + 1]
            frame_colors[2 * b : 2 * b + 2, :] = color

        update = scene.update_instanced_mesh("butterflies", frame_positions, frame_rotations, frame_colors)
        animation.append(update)

    num_frames = 300
    cameras = sp.Camera.orbit(num_frames, 3, 2)
    for i, camera in enumerate(cameras):
        frame = canvas.create_frame()
        frame.add_mesh(animation[i % num_anim_frames])
        frame.camera = camera


def test_camera_movement(asset):
    scene = sp.Scene()
    spin_canvas = scene.create_canvas_3d("spin")
    spiral_canvas = scene.create_canvas_3d("spiral")

    polar_bear = scene.create_image(image_id="polar_bear")
    polar_bear.load(asset("PolarBear.png"))
    uv_texture = scene.create_image(image_id="texture")
    uv_texture.load(asset("uv.png"))

    cube = scene.create_mesh("cube", texture_id=polar_bear.image_id)
    cube.add_cube()
    sphere = scene.create_mesh("sphere", texture_id=uv_texture.image_id)
    sphere.add_icosphere(steps=4, transform=sp.Transforms.translate([0, 1, 0]))

    spin_frustum = scene.create_mesh("spin_frustum", shared_color=sp.Colors.Red)
    spiral_frustum = scene.create_mesh("spiral_frustum", shared_color=sp.Colors.Green)

    spin_frustum.add_camera_frustum(depth=-1, transform=sp.Transforms.rotation_about_z(np.pi))
    spiral_frustum.add_camera_frustum(depth=-1, transform=sp.Transforms.rotation_about_z(np.pi))

    num_frames = 60
    for i in range(num_frames):
        spin_frame = spin_canvas.create_frame()
        spin_frame.add_meshes([cube, sphere])

        angle = i * np.pi * 2 / num_frames

        rotation = sp.Transforms.rotation_about_z(angle)
        spin_frame.camera = sp.Camera(center=[0, 0, 4], rotation=rotation, fov_y_degrees=30.0)

        spiral_frame = spiral_canvas.create_frame()
        spiral_frame.add_meshes([cube, sphere])

        camera_center = [4 * np.cos(angle), i * 4 / num_frames - 2, 4 * np.sin(angle)]
        spiral_frame.camera = sp.Camera(camera_center, look_at=[0, 0.5, 0])

        spiral_frame.add_mesh(spiral_frustum, transform=spiral_frame.camera.camera_to_world)
        spiral_frame.add_mesh(spin_frustum, transform=spin_frame.camera.camera_to_world)
        spin_frame.add_mesh(spiral_frustum, transform=spiral_frame.camera.camera_to_world)
        spin_frame.add_mesh(spin_frustum, transform=spin_frame.camera.camera_to_world)

    scene.link_canvas_events(spin_canvas, spiral_canvas)


def _set_audio(scene, canvas, path):
    audio = scene.create_audio()
    audio.load(path)
    canvas.media_id = audio.audio_id


def test_audio_tracks(asset):
    scene = sp.Scene()

    names = ["red", "green", "blue"]
    colors = [sp.Colors.Red, sp.Colors.Green, sp.Colors.Blue]
    frequencies = [0, 1, 0.5]

    graph = scene.create_graph("graph", width=900, height=150)
    for name, color, frequency in zip(names, colors, frequencies):
        mesh = scene.create_mesh()
        mesh.add_cube(color)
        canvas = scene.create_canvas_3d(name, width=300, height=300)
        _set_audio(scene, canvas, asset(name + ".ogg"))
        values = []

        for j in range(60):
            frame = canvas.create_frame()
            scale = math.sin(j * 2 * math.pi * frequency / 30)
            frame.add_mesh(mesh, sp.Transforms.scale((scale + 1) / 2 + 0.5))
            values.append(scale)

        graph.add_sparkline(name, values, color)

    names.append("graph")
    scene.grid("600px", "1fr auto", "1fr 1fr 1fr")
    scene.place("graph", "2", "1 / span 3")
    scene.link_canvas_events(*names)


def _angle_to_pos(angle, radius):
    return np.cos(angle) * radius + 200, np.sin(angle) * radius + 200


def test_circles_video(asset):
    scene = sp.Scene()

    video = scene.create_video()
    video.load(asset("circles.mp4"))

    tracking = scene.create_canvas_2d("tracking", background_color=sp.Colors.White)
    tracking.media_id = video.video_id

    multi = scene.create_canvas_2d("multi", background_color=sp.Colors.White)
    multi.media_id = video.video_id

    angles = np.linspace(0, 2 * np.pi, 360, endpoint=False)
    for angle in angles:
        frame = tracking.create_frame()
        frame.add_video(layer_id="video")

        red_pos = _angle_to_pos(angle, 160)
        frame.add_rectangle(red_pos[0] - 11, red_pos[1] - 11, 22, 22, [255, 0, 0], 2, layer_id="rect")
        frame.add_circle(red_pos[0], red_pos[1], 10, fill_color=[255, 0, 0], layer_id="dot")

        green_pos = _angle_to_pos(-2 * angle, 80)
        frame.add_rectangle(green_pos[0] - 11, green_pos[1] - 11, 22, 22, [0, 255, 0], 2, layer_id="rect")
        frame.add_circle(green_pos[0], green_pos[1], 10, fill_color=[0, 255, 0], layer_id="dot")

        blue_pos = _angle_to_pos(4 * angle, 40)
        frame.add_rectangle(blue_pos[0] - 11, blue_pos[1] - 11, 22, 22, [0, 0, 255], 2, layer_id="rect")
        frame.add_circle(blue_pos[0], blue_pos[1], 10, fill_color=[0, 0, 255], layer_id="dot")

        frame = multi.create_frame()
        frame.add_video("manual", red_pos[0] - 40, red_pos[1] - 40, 0.2, layer_id="red")
        frame.add_video("manual", green_pos[0] - 25, green_pos[1] - 25, 0.125, layer_id="green")
        frame.add_video("manual", 160, 160, 0.2, layer_id="blue")

    tracking.set_layer_settings({
        "rect": {"render_order": 0},
        "video": {"render_order": 1},
        "dot": {"render_order": 2}
    })

    multi.set_layer_settings({
        "red": {},
        "green": {},
        "blue": {}
    })

    scene.link_canvas_events("tracking", "multi")


def _load_camera(camera_info):
    location = np.array(camera_info["location"], np.float32)
    euler_angles = np.array(camera_info["rotation"], np.float32)
    rotation = sp.Transforms.euler_angles_to_matrix(euler_angles, "XYZ")
    translation = sp.Transforms.translate(location)
    extrinsics = translation @ rotation
    world_to_camera = sp.Transforms.gl_world_to_camera(extrinsics)
    aspect_ratio = camera_info["width"] / camera_info["height"]
    projection = sp.Transforms.gl_projection(camera_info["fov"], aspect_ratio, 0.01, 100)

    return sp.Camera(world_to_camera, projection)


def _load_cameras(path):
    with open(path) as file:
        cameras = json.load(file)
        return [_load_camera(cameras[key])
                for key in cameras]


def test_multiview(asset):
    scene = sp.Scene()

    cameras = _load_cameras(asset("cameras.json"))

    texture = scene.create_image("texture")
    texture.load(asset("PolarBear.png"))
    cube = scene.create_mesh("cube")
    cube.texture_id = texture.image_id
    cube.add_cube(transform=sp.Transforms.scale(2))

    frustums = scene.create_mesh("frustums", layer_id="frustums")
    colors = [sp.Colors.Red, sp.Colors.Green, sp.Colors.Blue]
    paths = [asset(name) for name in ["render0.png", "render1.png", "render2.png"]]
    camera_images = []
    images = []

    for i, (color, path, camera) in enumerate(zip(colors, paths, cameras)):
        image = scene.create_image(path)
        image.load(path)
        frustums.add_camera_frustum(camera, color)

        image_mesh = scene.create_mesh("image{}".format(i),
                                       layer_id="images",
                                       shared_color=sp.Colors.Gray,
                                       double_sided=True,
                                       texture_id=image.image_id)
        image_mesh.add_camera_image(camera)

        images.append(image)
        camera_images.append(image_mesh)

    width = 640
    for i, camera in enumerate(cameras):
        height = width / camera.aspect_ratio
        canvas = scene.create_canvas_3d("hand{}".format(i), width, height, camera=camera)
        frame = canvas.create_frame()
        frame.add_mesh(cube)
        frame.add_mesh(frustums)
        frame.camera = camera
        for cam_mesh in camera_images:
            frame.add_mesh(cam_mesh)


def test_fading():
    scene = sp.Scene()

    cube = scene.create_mesh(layer_id="cube")
    cube.add_cube(sp.Colors.Green)

    sphere = scene.create_mesh(layer_id="sphere")
    sphere.add_sphere(sp.Colors.Red)

    canvas = scene.create_canvas_3d()
    for i in range(60):
        sphere_opacity = i / 59
        cube_opacity = 1 - sphere_opacity
        frame = canvas.create_frame()
        frame.add_mesh(cube)
        frame.add_mesh(sphere)
        frame.set_layer_settings({
            "cube": {"opacity": cube_opacity},
            "sphere": {"opacity": sphere_opacity}
        })



"""Module which extends the scenepic Mesh type."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import numpy as np

from ._scenepic import Mesh, MeshUpdate


class VertexBuffer:
    """Class which provides dictionary access to vertex buffer blocks.

    Description:
        The different parts of the vertex buffer can be accessed via a dictionary
        interface:

        - "pos": the vertex positions
        - "norm": the vertex normals
        - "rgb": the vertex colors (if present)
        - "uv": the vertex uvs (if present)

    Args:
        values (np.ndarray): The raw vertex buffer values
    """

    def __init__(self, values: np.ndarray):
        """Initializer."""
        self._values = values
        self._lookup = {
            "pos": slice(0, 3),
            "norm": slice(3, 6),
            "rgb": slice(6, 9),
            "uv": slice(6, 8)
        }

    @property
    def shape(self) -> tuple:
        """Shape of the entire vertex buffer."""
        return self._values.shape

    def copy(self) -> "VertexBuffer":
        """Returns a copy of the vertex buffer."""
        return VertexBuffer(self._values.copy())

    def __repr__(self) -> str:
        """Return a string representation of the vertex buffer."""
        return str(self._values)

    def __getitem__(self, key: str) -> np.ndarray:
        """Returns a sub-section of the buffer given the key.

        Args:
            key (str): one of ["pos", "norm", "rgb", "uv"]

        Returns:
            np.ndarray: a subsection of the buffer
        """
        assert key in self._lookup, "Invalid vertex buffer key " + key
        return self._values[:, self._lookup[key]]

    def __setitem__(self, key: str, data: np.ndarray):
        """Sets a subsection of the buffer specified by the key.

        Args:
            key (str): one of ["pos", "norm", "rgb", "uv"]
            data (np.ndarray): The new values to place in the buffer
        """
        assert key in self._lookup, "Invalid vertex buffer key " + key
        self._values[:, self._lookup[key]] = data


def vertex_buffer(self):
    """VertexBuffer: a reference to the vertex buffer."""
    return VertexBuffer(self.get_vertex_buffer())


def quantize(self, keyframe_index: int, fixed_point_range: float, keyframe_vertex_buffer: VertexBuffer):
    """Quantize the mesh update.

    Args:
        self (MeshUpdate): self reference
        keyframe_index (int): Index of the keyframe to use in quantizing this update
        fixed_point_range (float): The range to use for the fixed point representation.
        keyframe_vertex_buffer (VertexBuffer): The keyframe vertex buffer
    """
    self.quantize_(keyframe_index, fixed_point_range, keyframe_vertex_buffer._values)


def difference_range(self, vertex_buffer: VertexBuffer) -> float:
    """Returns the absolute range of values in the difference between this update and the buffer.

    Args:
        self (MeshUpdate): self reference
        vertex_buffer (VertexBuffer): the buffer to use in the comparison

    Return:
        float: the absolute range (from minimum to maximum) in the per-index difference between
               this update and the reference.
    """
    self.difference_range_(vertex_buffer._values)


Mesh.vertex_buffer = property(vertex_buffer)
MeshUpdate.vertex_buffer = property(vertex_buffer)
MeshUpdate.quantize = quantize
MeshUpdate.difference_range = difference_range



"""Module which extends the Camera type."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from ._scenepic import (
    Camera,
    camera_orbit
)

Camera.orbit = camera_orbit



"""Test the loop subdivision stencil."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import scenepic as sp


def test_loop_subdivision_stencil(asset, assert_json_equal):
    hand_lo = sp.load_obj(asset("hand.obj"))

    hand_verts1, hand_tris1 = sp.LoopSubdivStencil(hand_lo.triangles,
                                                   steps=1,
                                                   project_to_limit=True).apply(hand_lo.positions)
    hand_hi2 = hand_lo.subdivide(2)

    scene = sp.Scene()
    hand_mesh1 = scene.create_mesh("hand1", shared_color=sp.Colors.Pink)
    hand_mesh1.add_mesh_without_normals(hand_verts1, hand_tris1)
    head_mesh2 = scene.create_mesh("hand2", shared_color=sp.Colors.Blue)
    head_mesh2.add_mesh(hand_hi2)

    assert_json_equal(scene.get_json(), "loop_subdivision_stencil")



"""Component that wraps FFMPEG to allow for the production of videos."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from subprocess import PIPE, Popen, STDOUT
from typing import Tuple

import cv2
import numpy as np


class VideoWriter:
    """The VideoWriter class.

    Description:
        Provides a straightforward means of producing videos by providing a lightweight wrapper
        around the commonly used FFMPEG command-line tool. Is best used as a context
        manager as shown in the example below.

    Example:
        Provided the ffmpeg executable is installed on the system, the user can create a
        video in the following way:

        .. code-block:: python

            import cv2
            import numpy as np
            from scenepic import VideoWriter

            with VideoWriter("example_video.mp4", (256, 256)) as video:
                angles = np.linspace(0, 2 * np.pi, 60, endpoint=False)
                for angle in angles:
                    x = int(np.cos(angle) * 64 + 128)
                    y = int(np.sin(angle) * 64 + 128)
                    video.clear_frame()
                    cv2.circle(video.frame, (x, y), 16, (0, 0, 255), -1)
                    video.write_frame()

        This will result in the following video:

        .. image:: ../../ci/assets/example_video.gif
    """

    def __init__(self,
                 output_path: str,
                 frame_size: Tuple[int, int],
                 quality: int = None,
                 ffmpeg_path="ffmpeg",
                 background_color=(0, 0, 0),
                 rgb=False,
                 audio_path=None,
                 codec="libx264",
                 framerate=30,
                 text="",
                 text_color=(1, 1, 0),
                 font_scale=1):
        """Constructor.

        Args:
            output_path (str): The path to the output video file
            frame_size (Tuple[int, int]): The (width, height) of the frame in pixels
            quality (int, optional): The 'q' or 'crf' value driving the quality of the encoding.
                Defaults to None.
            ffmpeg_path (str, optional): The path to the FFMPEG executable. Default "ffmpeg".
            background_color (Tuple[float, float, float], optional): The (r, g, b) background
                color for the frame where 0.0 <= val <= 1.0. Defaults to (0, 0, 0).
            rgb (bool, optional): Whether the pixels are in RGB or BGR format. Defaults to False (bgr).
            audio_path (str, optional): Path to an audio file to demux with the video. Defaults to None.
            codec (str, optional): Codec to use for encoding the video. Defaults to "libx264".
            framerate (float, optional): Framerate for video
            text (str, optional): Text to burn into the frame
            text_color (Tuple[float, float, float], optional): Color for the text
            font_scale (float, optional): Scaling factor for the font.
                                          Defaults to 1.
        """

        self._output_path = output_path
        if codec == "libx264":
            self._codec_flags = ["-c:v", "libx264"]
            if quality is not None:
                self._codec_flags.extend(["-crf", str(quality)])
        else:
            self._codec_flags = ["-c:v", codec]
            if quality is not None:
                self._codec_flags.extend(["q", str(quality)])

        if audio_path:
            self._audio_flags = ["-i", audio_path]
        else:
            self._audio_flags = []

        self._framerate = framerate
        self._rgb = rgb
        self._frame_size = frame_size
        self._ffmpeg_path = ffmpeg_path
        self._process = None
        self._text = text
        self._frame = np.zeros((frame_size[1], frame_size[0], 3), np.uint8)

        self._background_color = self._to_color(background_color)
        self._text_color = self._to_color(text_color)
        self._font_scale = font_scale

    def _to_color(self, color: Tuple[float, float, float]) -> np.array:
        r, g, b = color
        assert 0 <= r <= 1 and 0 <= g <= 1 and 0 <= b <= 1
        if self._rgb:
            color = np.array([r, g, b])
        else:
            color = np.array([b, g, r])

        return (color * 255).astype(np.uint8)

    def __enter__(self):
        """Called when using the VideoWriter as a context manager."""
        self.start()
        return self

    def __exit__(self, type, value, traceback):  # noqa: A002
        """Called when using the VideoWriter as a context manager."""
        self.stop()
        return False

    def start(self):
        """Starts the video writing process.

        No need to call this method manually if using VideoWriter as a context manager.
        """
        self._process = Popen([self._ffmpeg_path, "-y", "-f", "image2pipe",
                               "-c:v", "ppm",
                               "-s", "{}x{}".format(*self._frame_size),
                               "-pix_fmt", "bgr24",
                               "-framerate", str(self._framerate),
                               "-i", "-",
                               *self._audio_flags,
                               *self._codec_flags,
                               "-pix_fmt", "yuv420p",
                               self._output_path],
                              stdin=PIPE, stderr=STDOUT)

    def stop(self):
        """Stops the video writing process and closes the video file.

        No need to call this method manually if using VideoWriter as a context manager.
        """
        self._process.stdin.flush()
        self._process.stdin.close()
        self._process.wait()

    @property
    def text(self) -> str:
        """Text to burn into the frame."""
        return self._text

    @text.setter
    def text(self, text: str):
        self._text = text

    @property
    def text_color(self) -> np.ndarray:
        """The color of the burned text."""
        return (self._text_color / 255).astype(np.float32)

    @text_color.setter
    def text_color(self, text_color: Tuple[float, float, float]):
        self._text_color = self._to_color(text_color)

    @property
    def frame(self) -> np.ndarray:
        """Returns the frame buffer as a (H, W, 3) uint8 numpy array."""
        return self._frame

    def clear_frame(self):
        """Clears the frame buffer, setting all pixels to the background color."""
        self._frame[:, :] = self._background_color

    def write_frame(self):
        """Write the frame buffer to the video."""
        if self.text:
            width, height = self._frame.shape[:2]
            size = min(width, height)
            font_scale = self._font_scale * size / 512
            thickness = max(1, int(2 * font_scale))
            self._frame = cv2.putText(self._frame, self._text,
                                      (50, height - 50),
                                      cv2.FONT_HERSHEY_SIMPLEX, font_scale,
                                      tuple(self._text_color.tolist()),
                                      thickness, cv2.LINE_AA)

        if self._rgb:
            cv2.cvtColor(self._frame, cv2.COLOR_RGB2BGR, self._frame)

        _, buffer = cv2.imencode(".PPM", self._frame)
        self._process.stdin.write(buffer)



"""Module which extends the scenepic Canvas3D type."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from typing import List, Union, Mapping

import numpy as np

from ._scenepic import (
    Camera,
    Canvas3D,
    FocusPoint,
    Frame3D,
    LayerSettings,
    Mesh,
    MeshUpdate
)


def create_frame(self,
                 frame_id="",
                 focus_point: FocusPoint = None,
                 meshes: List[Union[Mesh, MeshUpdate, str]] = None,
                 camera: Camera = None) -> Frame3D:
    """Creates a new Frame3D object and appends to the list of Frame3Ds in the Canvas3D.

    Args:
        frame_id (str): a unique identifier for the Frame3D (will be
                        automatically populated if not provided).
        focus_point (FocusPoint): optional 3D focus point for this frame (with optional 3D
                                  rotation for orientation lock), used in the ScenePic user interface.
        meshes (List[Union[Mesh, MeshUpdate, str]]): optionally pre-populate this Frame3D
                                                     with a set of Meshes, for convenience.
        camera (Camera): optionally set a camera for this frame

    Returns:
        Frame3D: a new Frame3D object
    """
    kwargs = {}
    if focus_point is not None:
        if isinstance(focus_point, (list, np.ndarray)):
            focus_point = FocusPoint(focus_point)

        kwargs["focus_point"] = focus_point

    if meshes is not None:
        kwargs["mesh_ids"] = [mesh if isinstance(mesh, str) else mesh.mesh_id
                              for mesh in meshes]

    if camera is not None:
        kwargs["camera"] = camera

    return self.create_frame_(frame_id, **kwargs)


def set_layer_settings(self, layer_settings: Mapping[str, Union[dict, LayerSettings]]):
    """Specify the visibilities and opacities of certain mesh layers.

    Description:
        Each Mesh object can optionally be part of a user-identified layer
        (see Mesh constructor).
        Calling set_layer_settings will result in an additional drop-down
        selector in the ScenePic user interface.

    Args:
        self (Canvas3D): the self reference
        layer_settings (Mapping[str, Union[dict, LayerSettings]]): a LayerSettings object
            or a dictionary. See LayerSettings for details.
    """
    updated_settings = {}
    for key, value in layer_settings.items():
        if isinstance(value, dict):
            if "visible" in value:
                print("`visible` keyword has been deprecated")
            else:
                updated_settings[key] = LayerSettings(**value)

    self.set_layer_settings_(updated_settings)


Canvas3D.create_frame = create_frame
Canvas3D.set_layer_settings = set_layer_settings



"""Module which extends the scenepic Frame3D type."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from typing import List, Mapping, Union

import numpy as np

from ._scenepic import Frame3D, LayerSettings, Mesh, MeshUpdate


def add_meshes(self, meshes: List[Union[Mesh, MeshUpdate, str]], transform: np.ndarray = None):
    """Adds meshes to the frame.

    Args:
        self (Frame3D): self reference
        meshes (List[Union[Mesh, MeshUpdate, str]]): A list of meshes and mesh updates (or their IDs)
        transform (np.ndarray, optional): an optional transform to apply to each mesh. Defaults to None.
    """
    mesh_ids = [mesh if isinstance(mesh, str) else mesh.mesh_id
                for mesh in meshes]

    if transform is None:
        self.add_meshes_(mesh_ids)
    else:
        self.add_meshes_(mesh_ids, transform)


def add_mesh(self, mesh: Union[Mesh, MeshUpdate, str], transform: np.ndarray = None):
    """Adds a mesh to the frame.

    Args:
        self (Frame3D): self reference
        mesh (Union[Mesh, MeshUpdate, str]): The mesh, mesh update, or mesh ID
        transform (np.ndarray, optional): An optional transform to apply to the mesh. Defaults to None.
    """
    mesh_id = mesh if isinstance(mesh, str) else mesh.mesh_id

    if transform is None:
        self.add_mesh_(mesh_id)
    else:
        self.add_mesh_(mesh_id, transform)


def set_layer_settings(self, layer_settings: Mapping[str, Union[dict, LayerSettings]]):
    """Specify the visibilities and opacities of certain mesh layers.

    Description:
        Each Mesh object can optionally be part of a user-identified layer
        (see Mesh constructor).
        Calling set_layer_settings will result in an additional drop-down
        selector in the ScenePic user interface.

    Args:
        self (Canvas3D): the self reference
        layer_settings (Mapping[str, Union[dict, LayerSettings]]): a LayerSettings object
            or a dictionary. See LayerSettings for details.
    """
    updated_settings = {}
    for key, value in layer_settings.items():
        if isinstance(value, dict):
            if "visible" in value:
                print("`visible` keyword has been deprecated")
            else:
                updated_settings[key] = LayerSettings(**value)

    self.set_layer_settings_(updated_settings)


Frame3D.add_meshes = add_meshes
Frame3D.add_mesh = add_mesh
Frame3D.set_layer_settings = set_layer_settings



"""Tests for transform calculations."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import numpy as np
import scenepic as sp


PI = np.array(np.pi, np.float32)
ANGLE = 60 * PI / 180
COS = np.cos(ANGLE)
SIN = np.sin(ANGLE)
VEC = np.array([0.1, 0.2, 0.3], np.float32)


def test_scale_1():
    expected = np.array([
        [0.1, 0, 0, 0],
        [0, 0.1, 0, 0],
        [0, 0, 0.1, 0],
        [0, 0, 0, 1]
    ], np.float32)

    np.testing.assert_allclose(sp.Transforms.scale(0.1), expected)


def test_scale_3():
    expected = np.array([
        [0.1, 0, 0, 0],
        [0, 0.2, 0, 0],
        [0, 0, 0.3, 0],
        [0, 0, 0, 1]
    ], np.float32)

    np.testing.assert_allclose(sp.Transforms.scale(VEC), expected)


def test_translate():
    expected = np.array([
        [1, 0, 0, 0.1],
        [0, 1, 0, 0.2],
        [0, 0, 1, 0.3],
        [0, 0, 0, 1]
    ], np.float32)

    np.testing.assert_allclose(sp.Transforms.translate(VEC), expected)


def test_rotation_about_x():
    expected = np.array([
        [1, 0, 0, 0],
        [0, COS, -SIN, 0],
        [0, SIN, COS, 0],
        [0, 0, 0, 1]
    ], np.float32)

    np.testing.assert_allclose(sp.Transforms.rotation_about_x(ANGLE), expected)


def test_rotation_about_y():
    expected = np.array([
        [COS, 0, SIN, 0],
        [0, 1, 0, 0],
        [-SIN, 0, COS, 0],
        [0, 0, 0, 1]
    ], np.float32)

    np.testing.assert_allclose(sp.Transforms.rotation_about_y(ANGLE), expected)


def test_rotation_about_z():
    expected = np.array([
        [COS, -SIN, 0, 0],
        [SIN, COS, 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, 1]
    ], np.float32)

    np.testing.assert_allclose(sp.Transforms.rotation_about_z(ANGLE), expected)


def test_euler_angles_to_matrix():
    expected = np.array([
        [0.936293, -0.289629, 0.198669, 0],
        [0.312992, 0.944703, -0.0978434, 0],
        [-0.159345, 0.153792, 0.97517, 0],
        [0, 0, 0, 1]
    ], np.float32)

    np.testing.assert_allclose(sp.Transforms.euler_angles_to_matrix(VEC), expected, rtol=0, atol=1e-6)


def test_rotation_to_align_x_to_axis():
    expected = np.array([
        [0.267261, -0.534522, -0.801784, 0],
        [0.534522, 0.774542, -0.338187, 0],
        [0.801784, -0.338187, 0.492719, 0],
        [0, 0, 0, 1]
    ], np.float32)

    np.testing.assert_allclose(sp.Transforms.rotation_to_align_x_to_axis(VEC), expected, rtol=0, atol=1e-6)


def test_quaternion_to_align_x_to_axis():
    expected = np.array([0, -0.503627, 0.335751, 0.796009], np.float32)

    np.testing.assert_allclose(sp.Transforms.quaternion_to_align_x_to_axis(VEC), expected, rtol=0, atol=1e-6)


def test_quaternion_to_matrix():
    quaternion = np.array([0, -0.503627, 0.335751, 0.796009], np.float32)
    expected = np.array([
        [0.267261, -0.534522, -0.801784, 0],
        [0.534522, 0.774542, -0.338187, 0],
        [0.801784, -0.338187, 0.492719, 0],
        [0, 0, 0, 1]
    ], np.float32)

    np.testing.assert_allclose(sp.Transforms.quaternion_to_matrix(quaternion), expected, rtol=0, atol=1e-5)


def test_gl_projection():
    fov_y_degrees = 45
    aspect_ratio = 16 / 9
    znear = 1
    zfar = 100
    expected = np.array([
        [1.357995, 0, 0, 0],
        [0, 2.414214, 0, 0],
        [0, 0, -1.020202, -2.020202],
        [0, 0, -1, 0],
    ], np.float32)

    actual = sp.Transforms.gl_projection(fov_y_degrees, aspect_ratio, znear, zfar)
    np.testing.assert_allclose(actual, expected, rtol=0, atol=1e-6)

    intrinsic = np.array([
        [7.5438633e+03, 1.1557838e+01, 2.0555000e+03],
        [0, 7.5612310e+03, 1.5035000e+03],
        [0, 0, 1]
    ], np.float32)
    width = 4112
    height = 3008
    expected = np.array([
        [3.6691942, -5.6215166e-03, 2.4319066e-04, 0],
        [0, 5.0274143, -3.3244680e-04, 0],
        [0, 0, -1.0202020, -2.0202019],
        [0, 0, -1, 0]
    ], np.float32)

    actual = sp.Transforms.gl_projection(intrinsic, width, height, znear, zfar)
    np.testing.assert_allclose(actual, expected, rtol=0, atol=1e-6)


def test_gl_world_to_camera():
    extrinsic = np.array([
        [0.6245465, 0.07013566, 0.77783203, -4.461637],
        [-0.12572001, -0.973942, 0.18876319, -0.7838287],
        [0.77080226, -0.21568045, -0.5994546, -4.4914584],
        [0, 0, 0, 1]
    ], np.float32)
    expected = np.array([
        [0.62454647, -0.12572001, 0.77080226, 6.149983],
        [-0.07013565, 0.9739419, 0.21568044, 1.4192035],
        [-0.77783203, -0.1887632, 0.5994546, -0.92593664],
        [0, 0, 0, 1]
    ], np.float32)

    np.testing.assert_allclose(sp.Transforms.gl_world_to_camera(extrinsic), expected, rtol=0, atol=1e-6)



"""Module which extends the ScenePic Scene type."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from typing import Union

from ._scenepic import (
    Canvas2D,
    Canvas3D,
    Graph,
    Image,
    js_lib_src,
    Mesh,
    Scene
)


def link_canvas_events(self, *canvases):
    """Specify that all input events should be linked across the provided set of Canvases.

    Args:
        self (Scene): self reference
        canvases: the canvases which should have linked events
    """
    ids = [canvas.canvas_id if isinstance(canvas, (Canvas3D, Canvas2D, Graph)) else canvas
           for canvas in canvases]
    self.link_canvas_events_(ids)


def create_mesh(self, mesh_id="", **kwargs) -> Mesh:
    """Create a new Mesh and append to the Scene.

    Description:
        Meshes are Scene-wide resources that can be reused across multiple
        Frames in multiple Canvas3Ds.

    Args:
        mesh_id (str, optional): a unique identifier for the Mesh (will be
                                 automatically populated if not provided). Defaults to None.
        layer_id (str, optional): an identifier to allow several Meshes to be grouped together
                                  into a "layer".  Layers should be used in conjunction with
                                  Canvas3D.set_layer_settings(). Defaults to None.
        double_sided (bool, optional): whether to turn off back face culling and draw the Mesh's
                                       triangles as double sided. Defaults to False.
        camera_space (bool, optional): whether this Mesh is defined in camera space (cannot
                                       be moved in the ScenePic user interface) or world space (standard).
                                       Defaults to False.
        shared_color (np.ndarray, optional): see `Mesh` constructor.
        texture_id (str, optional): see `Mesh` constructor.
        nn_texture (bool, optional): whether the shader should use nearest-neighbor
                    texture interpolation. Defaults to True.
        use_texture_alpha (bool, optional): whether to use the alpha channel in the texture
                                            for transparency (only relevant for textured Meshes).
                                            Defaults to True.
        is_billboard (bool, optional): draw this Mesh as a billboard (i.e. always facing the user)
                                       rather than rotating with the rest of the world. Defaults to False.
        is_label (bool, optional): this mesh will be treated specially as a label. Not for public use.
                                   Defaults to False.

    Returns:
        Mesh: a reference to the new Mesh object
    """
    if "texture_id" in kwargs:
        if isinstance(kwargs["texture_id"], Image):
            kwargs["texture_id"] = kwargs["texture_id"].image_id

    return self.create_mesh_(mesh_id, **kwargs)


def simulate_key_presses(self, canvas_id: Union[Canvas3D, Canvas2D, str], *keypresses):
    """Simulates one or more keypresses to fire off events in the ScenePic UI at startup, e.g. video playback.

    Args:
        self (Scene): self reference
        canvas_id (Union[Canvas3D, Canvas2D, str]): ID of the canvas on which to fire off this
                                                    event.
        keypresses: the key values
    """
    if isinstance(canvas_id, (Canvas2D, Canvas3D)):
        canvas_id = canvas_id.canvas_id

    self.simulate_key_presses_(canvas_id, keypresses)


def _repr_javascript_(self):
    """Internal method to enable Jupyter integration."""
    if self.script_cleared:
        raise Exception("You should not call clear_script() on Scenes that "
                        "you wish to use in Jupyter.")

    script = """
{};
window.Element = element;
{}
""".format(self.get_script(), js_lib_src())
    return script


Scene.link_canvas_events = link_canvas_events
Scene.create_mesh = create_mesh
Scene.simulate_key_presses = simulate_key_presses
Scene._repr_javascript_ = _repr_javascript_



"""Setup script for scenepic."""

from distutils.version import LooseVersion
import os
import platform
import re
import subprocess
import sys

from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext


REQUIRES = [
    "numpy>=1.16.2",
    "pillow>=5.2.0",
    "scipy>=1.4.1"
]

REQUIRES_VIDEO = [
    "opencv-python"
]

REQUIRES_DEV = [
    "azure-storage-blob>=12.0.0",
    "azure-common",
    "nbstripout",
    "pytest-azurepipelines",
    "pytest-cov",
    "pytest",
    "requests",
    "sphinx",
    "sphinx-autodoc-typehints",
    "sphinx-rtd-theme",
    "twine",
    "wheel"
]

LONG_DESCRIPTION = """
All platforms have good support for 2D images, with well-recognized formats
such as PNG and JPEG that can be viewed out of the box (no installation)
and shared trivially.

However, while many formats exist for 3D data, none are well-supported
without installation of tools such as MeshLab, Blender, etc.

ScenePic was created for 3D computer vision researchers such as those
working on [HoloLens](https://www.microsoft.com/en-gb/hololens)
and [Mesh](https://www.microsoft.com/en-us/mesh) at Microsoft.
It was designed to be a lightweight, reuseable 3D visualization
library, with the following desiderata in mind:

- Make experimentation with 3D data near effortless
- Incredibly easy to create and share 3D results
  * zero-install sharing of detailed 3D results using HTML
  * based on modern web standards so usable with any modern browser
    (tested in Edge, FireFox and Chrome)
  * embeddable in other HTML documents
- Performant
  * based on WebGL
- High quality visuals
- Works both offline or interactively in client-server setup
- Simple, clean API
  * friendly Python front-end
  * basic mesh json file format
  * other language front ends easy to add
"""

with open("VERSION", "r") as file:
    VERSION = file.read()


class CMakeExtension(Extension):
    """Extension class for cmake."""

    def __init__(self, name: str, source_dir=""):
        """Initializer.

        Args:
            name (str): The extension name
            source_dir (str, optional): The directory containing the source files. Defaults to "".
        """
        Extension.__init__(self, name, sources=[])
        self.source_dir = os.path.abspath(source_dir)


class CMakeBuild(build_ext):
    """Build driver for cmake."""

    def run(self):
        """Runs the build extension."""
        try:
            out = subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: "
                               + ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(re.search(r"version\s*([\d.]+)", out.decode()).group(1))
            if cmake_version < "3.1.0":
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext: CMakeExtension):
        """Builds the specified extension.

        Args:
            ext (CMakeExtension): the cmake extension.
        """
        cfg = "Debug" if self.debug else "Release"
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ["-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + extdir,
                      "-DCMAKE_BUILD_TYPE=" + cfg,
                      "-DPYTHON_EXECUTABLE=" + sys.executable,
                      "-DSCENEPIC_BUILD_PYTHON=1"]

        build_args = ["--config", cfg, "--target", "_scenepic"]

        if platform.system() == "Windows":
            cmake_args += ["-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir)]
            if platform.architecture()[0] == "64bit":
                cmake_args += ["-A", "x64"]
            else:
                cmake_args += ["-A", "Win32"]

            build_args += ["--", "/m"]
        else:
            cmake_args += ["-GUnix Makefiles", "-DCMAKE_BUILD_TYPE=" + cfg]
            build_args += ["--", "-j2"]

        env = os.environ.copy()
        env["CXXFLAGS"] = """{} -DVERSION_INFO="{}" """.format(env.get("CXXFLAGS", ""),
                                                               self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        subprocess.check_call(["cmake", ext.source_dir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(["cmake", "--build", "."] + build_args, cwd=self.build_temp)


setup(
    name="scenepic",
    version=VERSION,
    author="ScenePic Team",
    author_email="scenepic@microsoft.com",
    description="3D Visualization Made Easy",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages("src"),
    package_dir={"": "src"},
    python_requires=">=3.6, <4",
    ext_modules=[CMakeExtension("scenepic._scenepic")],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Visualization",
        "License :: OSI Approved :: MIT License"
    ],
    project_urls={
        "Documentation": "https://microsoft.github.io/scenepic/",
        "Bug Reports": "https://github.com/microsoft/scenepic/issues",
        "Source": "https://github.com/microsoft/scenepic",
    },
    url="https://microsoft.github.io/scenepic/",
    install_requires=REQUIRES,
    extras_require={
        "dev": REQUIRES_DEV,
        "video": REQUIRES_VIDEO
    },
    tests_require=["pytest"],
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False,
)



"""Tests for command serialization."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import numpy as np
from PIL import Image
import pytest
import scenepic as sp


def test_color():
    expected = sp.Color(0.5, 0, 0.2)
    actual = sp.ColorFromBytes(255, 0, 102) * 0.5
    np.testing.assert_array_almost_equal(actual, expected)


def test_camera(assert_json_equal):
    center = np.array([0, 2, 0], np.float32)
    look_at = np.array([0, 1, 0], np.float32)
    up_dir = np.array([1, 0, 0], np.float32)
    fov_y_degrees = 45.0
    initial_aspect_ratio = 1.5
    new_aspect_ratio = 0.9
    znear = 0.01
    zfar = 20

    rotation = np.array([
        [0, 0, 1, 0],
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 0, 1]
    ], np.float32)

    world_to_camera = np.array([
        [0, 0, 1, 0],
        [1, 0, 0, 0],
        [0, 1, 0, -2],
        [0, 0, 0, 1]
    ], np.float32)

    camera_to_world = np.array([
        [0, 1, 0, 0],
        [0, 0, 1, 2],
        [1, 0, 0, 0],
        [0, 0, 0, 1]
    ], np.float32)

    projection = sp.Transforms.gl_projection(fov_y_degrees, initial_aspect_ratio, znear, zfar)

    look_at_cam = sp.Camera(center, look_at, up_dir, fov_y_degrees, znear, zfar, initial_aspect_ratio)
    np.testing.assert_array_almost_equal(look_at_cam.center, center)
    np.testing.assert_array_almost_equal(look_at_cam.look_at, look_at)
    np.testing.assert_array_almost_equal(look_at_cam.up_dir, up_dir)
    np.testing.assert_array_almost_equal(look_at_cam.world_to_camera, world_to_camera)
    np.testing.assert_array_almost_equal(look_at_cam.camera_to_world, camera_to_world)
    np.testing.assert_array_almost_equal(look_at_cam.projection, projection)
    assert_json_equal(str(look_at_cam), "camera")

    look_at_cam.aspect_ratio = new_aspect_ratio
    assert look_at_cam.aspect_ratio == pytest.approx(new_aspect_ratio)

    rt_cam = sp.Camera(center, rotation, fov_y_degrees, znear, zfar, new_aspect_ratio)
    np.testing.assert_array_almost_equal(rt_cam.center, center)
    np.testing.assert_array_almost_equal(rt_cam.look_at, look_at)
    np.testing.assert_array_almost_equal(rt_cam.up_dir, up_dir)
    np.testing.assert_array_almost_equal(rt_cam.world_to_camera, world_to_camera)
    np.testing.assert_array_almost_equal(rt_cam.camera_to_world, camera_to_world)
    assert rt_cam.aspect_ratio == pytest.approx(new_aspect_ratio)
    rt_cam.aspect_ratio = initial_aspect_ratio
    np.testing.assert_array_almost_equal(rt_cam.projection, projection)
    assert_json_equal(str(rt_cam), "camera")

    cam = sp.Camera(world_to_camera, fov_y_degrees, znear, zfar, new_aspect_ratio)
    np.testing.assert_array_almost_equal(cam.center, center)
    np.testing.assert_array_almost_equal(cam.look_at, look_at)
    np.testing.assert_array_almost_equal(cam.up_dir, up_dir)
    np.testing.assert_array_almost_equal(cam.world_to_camera, world_to_camera)
    np.testing.assert_array_almost_equal(cam.camera_to_world, camera_to_world)
    assert cam.aspect_ratio == pytest.approx(new_aspect_ratio)
    cam.aspect_ratio = initial_aspect_ratio
    np.testing.assert_array_almost_equal(cam.projection, projection)
    assert_json_equal(str(cam), "camera")


def test_canvas2d(assert_json_equal):
    scene = sp.Scene()
    canvas2d = scene.create_canvas_2d("canvas2d")
    frame2d = canvas2d.create_frame()
    frame2d.add_circle(0, 0, 5)

    frame2d = canvas2d.create_frame()
    frame2d.add_rectangle(5, 6, 7, 8)

    frame2d = canvas2d.create_frame()
    frame2d.add_text("test", 1, 1)

    frame2d = canvas2d.create_frame()
    frame2d.add_image("rand")

    assert_json_equal(str(canvas2d), "canvas2d")
    canvas2d.clear_script()
    frame2d = canvas2d.create_frame()
    positions = np.array(
        [[0, 0],
         [1, 1],
         [2, 2]], np.float32
    )
    frame2d.add_line(positions)

    assert_json_equal(str(canvas2d), "canvas2d_cleared")


def test_canvas3d(assert_json_equal, color):
    scene = sp.Scene()

    cube_mesh = scene.create_mesh("cube")
    cube_mesh.add_cube(color)

    cone_mesh = scene.create_mesh("cone")
    cone_mesh.add_cone(color)

    disc_mesh = scene.create_mesh("disc")
    disc_mesh.add_disc(color)

    icosphere_mesh = scene.create_mesh("icosphere")
    icosphere_mesh.add_icosphere(color)

    cylinder_mesh = scene.create_mesh("cylinder")
    cylinder_mesh.add_cylinder(color)

    canvas3d = scene.create_canvas_3d("canvas3d")
    frame3d = canvas3d.create_frame("", [1, 0, 0])
    frame3d.add_mesh(cube_mesh)

    frame3d = canvas3d.create_frame()
    frame3d.add_mesh(disc_mesh, sp.Transforms.Scale(5))

    frame3d = canvas3d.create_frame()
    frame3d.add_mesh(icosphere_mesh)

    frame3d = canvas3d.create_frame()
    frame3d.add_mesh(cylinder_mesh)

    assert_json_equal(str(canvas3d), "canvas3d")

    canvas3d.clear_script()
    frame3d = canvas3d.create_frame()
    frame3d.add_mesh(cone_mesh)

    assert_json_equal(str(canvas3d), "canvas3d_cleared")


def test_drop_down_menu(assert_json_equal):
    scene = sp.Scene()
    drop_down_menu = scene.create_drop_down_menu("", "DropDown")
    drop_down_menu.items = ["one", "two", "three"]
    assert_json_equal(str(drop_down_menu), "drop_down_menu")


def test_frame2d(assert_json_equal):
    scene = sp.Scene()
    canvas2d = scene.create_canvas_2d()
    frame2d = canvas2d.create_frame()
    frame2d.add_circle(0, 0, 5.0)
    assert_json_equal(str(frame2d), "frame2d")


def test_frame3d(assert_json_equal, color):
    scene = sp.Scene()
    cube_mesh = scene.create_mesh("cube")
    cube_mesh.add_cube(color)

    canvas3d = scene.create_canvas_3d()
    frame3d = canvas3d.create_frame("", [1, 0, 0])
    frame3d.add_mesh(cube_mesh)

    assert_json_equal(str(frame3d), "frame3d")


def test_image(assert_json_equal, color, asset):
    scene = sp.Scene()
    image = scene.create_image("rand")
    image.load(asset("rand.png"))
    assert_json_equal(str(image), "image")

    pixels = np.array(Image.open(asset("rand.png")))
    image.from_numpy(pixels, "png")
    assert_json_equal(str(image), "image")

    with open(asset("rand.png"), "rb") as file:
        buffer = file.read()

    image.load_from_buffer(buffer, "png")
    assert_json_equal(str(image), "image")

    mesh = scene.create_mesh("image")
    mesh.texture_id = image.image_id
    mesh.add_image()
    assert_json_equal(str(mesh), "image_mesh")


def test_audio(assert_json_equal, asset):
    scene = sp.Scene()
    audio = scene.create_audio("hello")
    audio.load(asset("hello.mp3"))
    assert_json_equal(str(audio), "audio")


def test_video(assert_json_equal, asset):
    scene = sp.Scene()
    video = scene.create_video("test")
    video.load(asset("test.mp4"))
    assert_json_equal(str(video), "video")


def test_label(assert_json_equal):
    scene = sp.Scene()
    scene.create_label()
    assert_json_equal(scene.get_json(), "label")


def test_primitives(assert_json_equal, asset, color):
    scene = sp.Scene()
    texture = scene.create_image("uv")
    texture.load(asset("uv.png"))

    mesh = scene.create_mesh("triangle")
    mesh.add_triangle(color)
    assert_json_equal(str(mesh), "triangle")

    mesh = scene.create_mesh("quad")
    mesh.add_quad(color)
    assert_json_equal(str(mesh), "quad")

    mesh = scene.create_mesh("cube")
    mesh.add_cube(color)
    assert_json_equal(str(mesh), "cube")

    mesh = scene.create_mesh("cube_texture", texture_id=texture.image_id)
    mesh.add_cube()
    assert_json_equal(str(mesh), "cube_texture")

    mesh = scene.create_mesh("thickline")
    mesh.add_thickline(color)
    assert_json_equal(str(mesh), "thickline")

    mesh = scene.create_mesh("cone")
    mesh.add_cone(color)
    assert_json_equal(str(mesh), "cone")

    mesh = scene.create_mesh("trunc_cone")
    mesh.add_cone(color, truncation_height=0.7)
    assert_json_equal(str(mesh), "trunc_cone")

    mesh = scene.create_mesh("coordinate_axes_0")
    mesh.add_coordinate_axes()
    assert_json_equal(str(mesh), "coordinate_axes_0")

    mesh = scene.create_mesh("coordinate_axes_1", shared_color=sp.Colors.White)
    mesh.add_cube()
    mesh.add_coordinate_axes()
    assert_json_equal(str(mesh), "coordinate_axes_1")

    mesh = scene.create_mesh("camera_frustum")
    mesh.add_camera_frustum(color)
    assert_json_equal(str(mesh), "camera_frustum")

    mesh = scene.create_mesh("disc")
    mesh.add_disc(color)
    assert_json_equal(str(mesh), "disc")

    mesh = scene.create_mesh("cylinder")
    mesh.add_cylinder(color)
    assert_json_equal(str(mesh), "cylinder")

    mesh = scene.create_mesh("sphere")
    mesh.add_sphere(color)
    assert_json_equal(str(mesh), "sphere")

    mesh = scene.create_mesh("icosphere")
    mesh.add_icosphere(color)
    assert_json_equal(str(mesh), "icosphere")

    mesh = scene.create_mesh("icosphere_texture", texture_id=texture.image_id)
    mesh.add_icosphere(steps=1)
    assert_json_equal(str(mesh), "icosphere_texture")

    mesh = scene.create_mesh("uv_sphere")
    mesh.add_uv_sphere(color)
    assert_json_equal(str(mesh), "uv_sphere")

    mesh = scene.create_mesh("point_cloud")
    positions = []
    for x in range(5):
        for y in range(5):
            for z in range(5):
                positions.append([x, y, z])
    positions = np.array(positions, np.float32)
    positions = (positions / 2) - 1
    mesh.add_cube(color)
    mesh.apply_transform(sp.Transforms.scale(0.01))
    mesh.enable_instancing(positions)
    assert_json_equal(str(mesh), "point_cloud")

    mesh = scene.create_mesh("line_cloud")
    mesh.add_lines(positions, positions * 10, color)
    assert_json_equal(str(mesh), "line_cloud")


def test_io(assert_json_equal, asset):
    scene = sp.Scene()

    image = scene.create_image("texture")
    image.load(asset("PolarBear.png"))

    mesh = scene.create_mesh("cube")
    mesh.texture_id = image.image_id

    mesh_info = sp.load_obj(asset("cube.obj"))
    mesh.add_mesh(mesh_info)

    assert_json_equal(str(mesh), "io")


def test_mesh_update(assert_json_equal, color):
    scene = sp.Scene()

    mesh = scene.create_mesh("base")
    mesh.add_triangle(color)

    positions = np.array([
        [0, 0, 0],
        [1, 0, 0],
        [0, 0, 1]
    ], np.float32)

    normals = np.array([
        [0, -1, 0],
        [0, -1, 0],
        [0, -1, 0]
    ], np.float32)

    colors = np.array([
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1]
    ], np.float32)

    update = scene.update_mesh_positions("base", positions, "update0")

    assert_json_equal(str(update), "update0")

    update = scene.update_mesh("base", positions, normals, colors, "update1")

    assert_json_equal(str(update), "update1")

    keyframe_buffer = update.vertex_buffer.copy()
    keyframe_buffer["pos"][0] = [0, 1, 1]

    update.quantize(1, 6.0, keyframe_buffer)

    assert_json_equal(str(update), "update_quantized")

    instance_pos = np.array([
        [0, 1, 2],
        [2, 0, 1],
        [1, 0, 2]
    ], np.float32)

    instance_rot = np.array([
        [0.11, 0.22, 0.46, 0.85],
        [0.46, -0.12, -0.22, 0.85],
        [0.22, -0.12, 0.46, 0.85]
    ], np.float32)

    instance_c = np.array([
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1]
    ], np.float32)

    mesh.enable_instancing(instance_pos, instance_rot, instance_c)

    instance_pos[0] = [1, 1, 0]

    update = scene.update_mesh_positions("base", instance_pos, "update2")

    assert_json_equal(str(update), "update2")

    instance_pos[1] = [1, 0, 1]
    instance_rot[0] = [0.24, 0.24, 0.06, 0.94]
    instance_c[0] = [0.5, 0, 0]

    update = scene.update_instanced_mesh("base", instance_pos, instance_rot, instance_c, "update3")

    assert_json_equal(str(update), "update3")


def test_quantization(assert_json_equal, color):
    scene = sp.Scene()

    mesh = scene.create_mesh("base")
    mesh.add_triangle(color)

    for i in range(20):
        positions = np.array([
            [0, 0, 0],
            [1, i * 0.05, 0],
            [0, 1, 0]
        ], np.float32)
        scene.update_mesh_positions("base", positions)

    info = scene.quantize_updates(1e-5)

    assert info["base"].keyframe_count == 2
    assert_json_equal(scene.get_json(), "quantization")


def _create_tetrahedron():
    vertices = np.array([
        [-0.5, -0.32476, -0.20412],
        [0.5, -0.32476, -0.20412],
        [0, 0.541266, -0.20412],
        [0, 0.108253, 0.612372]
    ], np.float32)

    triangles = np.array([
        [0, 1, 3],
        [1, 2, 3],
        [2, 0, 3],
        [0, 2, 1]
    ], np.uint32)

    return vertices, triangles


SIZE = 500


def test_scene(assert_json_equal, asset):
    scene = sp.Scene("test")
    image = scene.create_image("rand")
    image.load(asset("rand.png"))
    mesh_rand = scene.create_mesh()
    mesh_rand.texture_id = image.image_id
    mesh_rand.layer_id = "Test"
    mesh_rand.add_image([-0.5, -0.5, 0], [2, 0, 0])

    tet_verts, tet_tris = _create_tetrahedron()

    model_mesh = scene.create_mesh()
    model_mesh.shared_color = sp.Color(1, 0, 0)
    model_mesh.add_mesh_without_normals(tet_verts, tet_tris)
    model_mesh.reverse_triangle_order()

    canvas_rand = scene.create_canvas_3d("", SIZE, SIZE)
    tet_center = tet_verts.mean(axis=0)
    canvas_tet = scene.create_canvas_3d("", SIZE, SIZE)
    canvas_tet.camera = sp.Camera(tet_center + np.array([0, 0, 0.5], np.float32),
                                  tet_center, [0, 1, 0], 45.0)
    canvas_tet.shading = sp.Shading(sp.Colors.White)
    canvas_tet.ui_parameters = sp.UIParameters()

    n_frames = 5
    for i in range(n_frames):
        angle = 2 * np.pi * i / n_frames

        frame_rand = canvas_rand.create_frame()
        frame_rand.add_mesh(mesh_rand, sp.Transforms.Translate([np.cos(angle), np.sin(angle), 0]))

        mesh_primitives = scene.create_mesh()
        mesh_primitives.layer_id = "Primitives"
        mesh_primitives.add_disc(sp.Color(0, 1, 0),
                                 sp.Transforms.Scale(0.2 + 0.2 * (1 + np.cos(angle))),
                                 10, False, True)
        mesh_primitives.add_cube(sp.Color(0, 0, 1), sp.Transforms.Translate([-1, -1, -3]))
        frame_rand.add_mesh(mesh_primitives)

        mesh_noise = scene.create_mesh()
        mesh_noise.shared_color = sp.Color(1, 0, 0)
        mesh_noise.layer_id = "Noise"
        mesh_noise.add_cylinder()
        mesh_noise.apply_transform(sp.Transforms.Scale([0.02, 0.1, 0.1]))
        mesh_noise.apply_transform(sp.Transforms.RotationToRotateXAxisToAlignWithAxis([0.5, 0.5, 0.5]))
        positions = np.zeros((16, 3), np.float32)
        for j in range(16 * 3):
            value = (j % 4) / 4.0
            positions[j // 3, j % 3] = 2 * value - 1

        mesh_noise.enable_instancing(positions)
        frame_rand.add_mesh(mesh_noise)

        frame_tet = canvas_tet.create_frame("", tet_center)
        frame_tet.add_mesh(model_mesh, sp.Transforms.RotationAboutZAxis(np.cos(angle)))

    canvas_rand.set_layer_settings({
        "Test": {"filled": True},
        "Primitives": {"filled": False},
        "Noise": {"filled": False, "opacity": 0.5}
    })

    assert_json_equal(scene.get_json(), "scene")

    scene.clear_script()

    frame_tet = canvas_tet.create_frame("", tet_center)
    frame_tet.add_mesh(model_mesh, sp.Transforms.RotationAboutZAxis(np.cos(2 * np.pi)))

    assert_json_equal(scene.get_json(), "scene_cleared")


def test_layer_settings(assert_json_equal):
    scene = sp.Scene()
    canvas3d = scene.create_canvas_3d()
    canvas2d = scene.create_canvas_2d()

    mesh0 = scene.create_mesh("null")
    mesh1 = scene.create_mesh("one", layer_id="one")
    mesh2 = scene.create_mesh("two", layer_id="two")
    image0 = scene.create_image("null")
    image1 = scene.create_image("one")
    image2 = scene.create_image("two")
    video = scene.create_video()

    canvas2d.media_id = video.video_id

    for _ in range(2):
        frame = canvas3d.create_frame()
        frame.add_mesh(mesh0)
        frame.add_mesh(mesh1)
        frame.add_mesh(mesh2)

        frame = canvas2d.create_frame()
        frame.add_circle(0, 0, 1)
        frame.add_circle(0, 0, 2, layer_id="one")
        frame.add_circle(0, 0, 3, layer_id="two")

        frame.add_text("null", 0, 0)
        frame.add_text("one", 100, 100, layer_id="one")
        frame.add_text("two", 200, 200, layer_id="two")

        coords = np.zeros((2, 2), np.float32)
        frame.add_line(coords)
        frame.add_line(coords + 1, layer_id="one")
        frame.add_line(coords + 2, layer_id="two")

        frame.add_rectangle(0, 0, 0, 0)
        frame.add_rectangle(0, 0, 10, 10, layer_id="one")
        frame.add_rectangle(0, 0, 20, 20, layer_id="two")

        frame.add_image(image0)
        frame.add_image(image1, layer_id="one")
        frame.add_image(image2, layer_id="two")

        frame.add_video("manual", 0, 0)
        frame.add_video("manual", 1, 1, layer_id="one")
        frame.add_video("manual", 2, 2, layer_id="two")

    layer_settings = {
        "one": {"opacity": 0.5, "render_order": 1},
        "two": {"opacity": 0.9}
    }

    canvas2d.set_layer_settings(layer_settings)
    canvas3d.set_layer_settings(layer_settings)

    assert_json_equal(str(canvas2d), "layer_settings_2d")
    assert_json_equal(str(canvas3d), "layer_settings_3d")


def test_shading(assert_json_equal):
    shading = sp.Shading()
    assert_json_equal(str(shading), "shading")


def test_text_panel(assert_json_equal):
    scene = sp.Scene()
    text_panel = scene.create_text_panel()
    text_panel.text = "Text"
    text_panel.title = "Title"
    assert_json_equal(str(text_panel), "text_panel")


def test_ui_parameters(assert_json_equal):
    ui_parameters = sp.UIParameters()
    assert_json_equal(str(ui_parameters), "ui_parameters")


def test_graph(assert_json_equal):
    scene = sp.Scene()
    graph = scene.create_graph()
    t = np.linspace(0, 1, 30)
    graph.add_sparkline("cos", np.cos(t), sp.Colors.Red)
    graph.add_sparkline("sin", np.sin(t), line_width=2)
    assert_json_equal(str(graph), "graph")



"""Module which extends the scenepic Frame2D type."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from typing import Union

import numpy as np

from ._scenepic import Frame2D, Image


def add_image(self, image_id: Union[Image, str], position_type="fit", **kwargs):
    """Add an image to the frame.

    Args:
        self (Frame2D): the self reference
        image_id (str): the unique identifier of the image
        position_type (str, optional): one of "fit", "fill", "stretch", or "manual". Defaults to "fit".
        x (float, optional): x coordinate to draw image at if position_type == "manual". Defaults to 0.0.
        y (float, optional): y coordinate to draw image at if position_type == "manual". Defaults to 0.0.
        scale (float, optional): scale factor to draw image by if position_type == "manual". Defaults to 1.0.
        smoothed (bool, optional): whether to draw an image with smoothing or using nearest neighbor. Defaults to False.
        layer_id (str, optional): Unique ID of the layer associated with this primitive. Defaults to "".
    """
    if isinstance(image_id, Image):
        image_id = image_id.image_id

    self.add_image_(image_id, position_type, **kwargs)


def add_circle(self, x: float, y: float, radius: float,
               line_color=None, line_width=1.0, fill_color=None,
               layer_id=""):
    """Add a circle to the frame.

    Args:
        self (Frame2D): the self reference
        x (float): the center x coordinate
        y (float): the center y coordinate
        radius (float): the radius of the circle
        line_color (np.ndarray, optional): the color of the line. Defaults to Black.
        line_width (float, optional): the width of the line. Defaults to 1.0.
        fill_color (np.ndarray, optional): the fill color to use. Defaults to None.
        layer_id (str, optional): Unique ID of the layer associated with this primitive. Defaults to "".
    """
    kwargs = {
        "line_width": line_width,
        "layer_id": layer_id
    }
    if line_color is not None:
        kwargs["line_color"] = np.array(line_color, np.float32)[:3]

    if fill_color is not None:
        kwargs["fill_color"] = np.array(fill_color, np.float32)[:3]

    self.add_circle_(x, y, radius, **kwargs)


def add_text(self, text: str, left: float, bottom: float,
             color=None, size_in_pixels=12.0, font_family="sans-serif",
             layer_id=""):
    """Add some text to the frame.

    Args:
        self (Frame2D): The self reference
        text (str): the text to display
        left (float): the pixel position of the left side of the text
        bottom (float): the pixel position of the bottom of the text
        color (np.ndarray, optional): the color of the text. Defaults to White.
        size_in_pixels (float, optional): the vertical size of the text in pixels. Defaults to 12.0.
        font_family (str, optional): the font to use for the text. Defaults to "sans-serif".
        layer_id (str, optional): Unique ID of the layer associated with this primitive. Defaults to "".
    """
    kwargs = {
        "size_in_pixels": size_in_pixels,
        "font_family": font_family,
        "layer_id": layer_id
    }

    if color is not None:
        kwargs["color"] = np.array(color, np.float32)[:3]

    self.add_text_(text, left, bottom, **kwargs)


Frame2D.add_image = add_image
Frame2D.add_circle = add_circle
Frame2D.add_text = add_text



"""Module which extends the scenepic Image type."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from io import BytesIO

import numpy as np
import PIL.Image

from ._scenepic import Image


def load_from_buffer(self, data: bytes, ext: str):
    """Load an image from a pre-existing data buffer and extension.

    Args:
        self (Image): self reference
        data (bytes): the binary data for the image
        ext (str): the format of the image data [e.g. "gif", "jpg", "png"]
    """
    self.data = list(data)
    self.ext = ext


def from_numpy(self, array: np.ndarray, format_str="png"):  # noqa: A002
    """Loads the image object from raw RGB values provided as a numpy array.

    Args:
        self (Image): self reference
        array (np.ndarray): The raw pixel RGB as either 0-255 bytes or 0-1 floats.
        format_str (str, optional): The format to use when compressing the image. Defaults to "png".
    """
    buff = BytesIO()
    if array.dtype.kind == "f":
        if array.max() <= 1.0:
            array *= 255

        array = array.astype(np.uint8)

    image = PIL.Image.fromarray(array)
    image.save(buff, format_str)
    self.data = list(buff.getvalue())
    self.ext = format_str
    buff.close()


def _data_get(self):
    return bytes(self.data_get_)


def _data_set(self, data):
    self.data_set_(list(data))


Image.load_from_buffer = load_from_buffer
Image.from_numpy = from_numpy
Image.data = property(_data_get, _data_set, doc="The binary image data")



"""Module which extends the scenepic Video type."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from ._scenepic import Video


def load_from_buffer(self, data: bytes, ext: str):
    """Load video from a pre-existing data buffer and extension.

    Args:
        self (Video): self reference
        data (bytes): the binary data for the video
        ext (str): the file extension (e.g. mp4, mkv)
    """
    self.data = list(data)
    self.ext = ext


def _data_get(self):
    return bytes(self.data_get_)


def _data_set(self, data):
    self.data_set_(list(data))


Video.load_from_buffer = load_from_buffer
Video.data = property(_data_get, _data_set, doc="The binary audio data")



"""ScenePic python interface module.

All platforms have good support for 2D images, with well-recognized formats
such as PNG and JPEG that can be viewed out of the box (no installation)
and shared trivially.

However, while many formats exist for 3D data, none are well-supported
without installation of tools such as MeshLab, Blender, etc.

ScenePic was created for 3D computer vision researchers such as those
working on [HoloLens](https://www.microsoft.com/en-gb/hololens)
and [Mesh](https://www.microsoft.com/en-us/mesh) at Microsoft.
It was designed to be a lightweight, reuseable 3D visualization
library, with the following desiderata in mind:

- Make experimentation with 3D data near effortless
- Incredibly easy to create and share 3D results
  * zero-install sharing of detailed 3D results using HTML
  * based on modern web standards so usable with any modern browser
    (tested in Edge, FireFox and Chrome)
  * embeddable in other HTML documents
- Performant
  * based on WebGL
- High quality visuals
- Works both offline or interactively in client-server setup
- Simple, clean API
  * friendly Python front-end
  * basic mesh json file format
  * other language front ends easy to add
"""

from typing import Tuple, Union

import numpy as np

from . import _scenepic
from ._scenepic import (
    ColorFromBytes,
    DropDownMenu,
    FocusPoint,
    Graph,
    js_lib_src,
    Label,
    LayerSettings,
    load_obj,
    LoopSubdivisionStencil,
    MeshInfo,
    Shading,
    TextPanel,
    UIParameters,
)
from .audio_track import AudioTrack
from .camera import Camera
from .canvas2d import Canvas2D
from .canvas3d import Canvas3D
from .frame2d import Frame2D
from .frame3d import Frame3D
from .image import Image
from .mesh import Mesh, MeshUpdate, VertexBuffer
from .scene import Scene
from .version import __version__
from .video import Video

try:
    from .video_writer import VideoWriter
except ImportError:
    print("scenepic[info]> Disabling VideoWriter due to ImportError")

__all__ = [
    "__version__",
    "AudioTrack",
    "Camera",
    "Canvas2D",
    "ColorFromBytes",
    "DropDownMenu",
    "FocusPoint",
    "Frame3D",
    "Graph",
    "js_lib_src",
    "MeshInfo",
    "Label",
    "LayerSettings",
    "load_obj",
    "Shading",
    "UIParameters",
    "Canvas3D",
    "Frame2D",
    "Image",
    "LoopSubdivisionStencil",
    "Mesh",
    "MeshUpdate",
    "Scene",
    "TextPanel",
    "VertexBuffer",
    "Video",
    "VideoWriter",
    "WebSocketHandler",
]


def Color(r: float, g: float, b: float) -> np.ndarray:  # noqa: N802
    """Convenience method for creating a numpy 3-vector from R,G,B float values.

    Args:
        r (float): red value [0,1].
        g (float): green value [0,1].
        b (float): blue value [0,1].

    Raises:
        ValueError: Will raise this if the values are out of range

    Returns:
        np.ndarray: a color array
    """
    if 0.0 <= r <= 1.0 and 0.0 <= g <= 1.0 and 0.0 <= b <= 1.0:
        return np.array([r, g, b], dtype=np.float32)

    raise ValueError("Color values range: [0,1]")


class Colors:
    """Class containing standard colors.

    Attributes:
        Black (np.ndarray): (0, 0, 0)
        White (np.ndarray): (255, 255, 255)
        Red (np.ndarray): (255, 0, 0)
        Maroon (np.ndarray): (128, 0, 0)
        Pink (np.ndarray): (255, 200, 220)
        Brown (np.ndarray): (170, 110, 40)
        Orange (np.ndarray): (255, 150, 0)
        Coral (np.ndarray): (255, 215, 180)
        Olive (np.ndarray): (128, 128, 0)
        Yellow (np.ndarray): (255, 235, 0)
        Beige (np.ndarray): (255, 250, 200)
        Lime (np.ndarray): (190, 255, 0)
        Green (np.ndarray): (0, 190, 0)
        Mint (np.ndarray): (170, 255, 195)
        Teal (np.ndarray): (0, 128, 128)
        Cyan (np.ndarray): (100, 255, 255)
        Navy (np.ndarray): (0, 0, 128)
        Blue (np.ndarray): (67, 133, 255)
        Purple (np.ndarray): (130, 0, 150)
        Lavender (np.ndarray): (230, 190, 255)
        Magenta (np.ndarray): (255, 0, 255)
        Gray (np.ndarray): (128, 128, 128)
    """

    Black = _scenepic.Black
    White = _scenepic.White
    Red = _scenepic.Red
    Maroon = _scenepic.Maroon
    Pink = _scenepic.Pink
    Brown = _scenepic.Brown
    Orange = _scenepic.Orange
    Coral = _scenepic.Coral
    Olive = _scenepic.Olive
    Yellow = _scenepic.Yellow
    Beige = _scenepic.Beige
    Lime = _scenepic.Lime
    Green = _scenepic.Green
    Mint = _scenepic.Mint
    Teal = _scenepic.Teal
    Cyan = _scenepic.Cyan
    Navy = _scenepic.Navy
    Blue = _scenepic.Blue
    Purple = _scenepic.Purple
    Lavender = _scenepic.Lavender
    Magenta = _scenepic.Magenta
    Gray = _scenepic.Gray
    Grey = _scenepic.Gray


class Transforms:
    """Static class with convenience methods for creating various 3D homogeneous transforms."""

    scale = _scenepic.scale

    @staticmethod
    def Scale(scale: Union[np.ndarray, float]):  # noqa: N802
        """Creates a 3D homogeneous scale matrix.

        Args:
            scale (Union[np.ndarray, float]): the axis-specific scale values

        Returns:
            np.ndarray: a 3D homogeneous transform matrix
        """
        return _scenepic.scale(scale)

    translate = _scenepic.translate

    @staticmethod
    def Translate(vec: np.ndarray):  # noqa: N802
        """Creates a 3D homogeneous scale matrix.

        Args:
            vec (np.ndarray): the movement vector

        Returns:
            np.ndarray: a 3D homogeneous transform matrix
        """
        return _scenepic.translate(vec)

    rotation_matrix_from_axis_angle = _scenepic.rotation_matrix_from_axis_angle

    @staticmethod
    def RotationMatrixFromAxisAngle(axis: np.ndarray, angle: float):  # noqa: N802
        """Creates a 3D homogeneous scale matrix from axis and angle.

        Args:
            axis (np.ndarray): the axis of rotation
            angle (float): the angle to rotate in radians

        Returns:
            np.ndarray: a 3D homogeneous transform matrix
        """
        return _scenepic.rotation_matrix_from_axis_angle(axis, angle)

    quaternion_from_axis_angle = _scenepic.quaternion_from_axis_angle

    @staticmethod
    def QuaternionFromAxisAngle(axis: np.ndarray, angle: float):  # noqa: N802
        """Convert rotation axis and angle to a quaternion.

        Args:
            axis (np.ndarray): the axis of rotation
            angle (float): the angle to rotate in radians

        Returns:
            np.ndarray: a 3D homgenous transform matrix
        """
        return _scenepic.quaternion_from_axis_angle(axis, angle)

    axis_angle_to_align_x_to_axis = _scenepic.axis_angle_to_align_x_to_axis

    @staticmethod
    def AxisAngleRotationToRotateXAxisToAlignWithAxis(axis: np.ndarray):  # noqa: N802
        """Find the axis and angle of rotation that will ensure the x-axis aligns with the provided axis.

        Note:
            One degree of freedom is underconstrained.

        Args:
            axis (np.ndarray): the axis to which to align the x-axis

        Returns:
            np.ndarray: a 3D homogeneous transform matrix
        """
        return _scenepic.axis_angle_to_align_x_to_axis(axis)

    quaternion_to_align_x_to_axis = _scenepic.quaternion_to_align_x_to_axis

    @staticmethod
    def QuaternionToRotateXAxisToAlignWithAxis(axis: np.ndarray):  # noqa: N802
        """Creates a quaternion rotation vector that will ensure the x-axis aligns with the provided axis.

        Note:
            One degree of freedom is underconstrained.

        Args:
            axis (np.ndarray): the axis to which to align the x-axis

        Returns:
            np.ndarray: a quaternion
        """
        return _scenepic.quaternion_to_align_x_to_axis(axis)

    rotation_to_align_x_to_axis = _scenepic.rotation_to_align_x_to_axis

    @staticmethod
    def RotationToRotateXAxisToAlignWithAxis(axis: np.ndarray):  # noqa: N802
        """Creates a 3D homogeneous rotation matrix that will ensure the x-axis aligns with the provided axis.

        Note:
            One degree of freedom is underconstrained.

        Args:
            axis (np.ndarray): the axis to which to align the x-axis

        Returns:
            np.ndarray: a 3D homogeneous transform matrix
        """
        return _scenepic.rotation_to_align_x_to_axis(axis)

    rotation_about_x = _scenepic.rotation_about_x

    @staticmethod
    def RotationAboutXAxis(angle: float):  # noqa: N802
        """Creates a 3D homogeneous rotation matrix about the x-axis.

        Args:
            angle (float): the angle to rotation in radians

        Returns:
            np.ndarray: a 3D homogeous transform matrix
        """
        return _scenepic.rotation_about_x(angle)

    rotation_about_y = _scenepic.rotation_about_y

    @staticmethod
    def RotationAboutYAxis(angle: float):  # noqa: N802
        """Creates a 3D homogeneous rotation matrix about the y-axis.

        Args:
            angle (float): the angle to rotation in radians

        Returns:
            np.ndarray: a 3D homogeous transform matrix
        """
        return _scenepic.rotation_about_y(angle)

    rotation_about_z = _scenepic.rotation_about_z

    @staticmethod
    def RotationAboutZAxis(angle: float):  # noqa: N802
        """Creates a 3D homogeneous rotation matrix about the z-axis.

        Args:
            angle (float): the angle to rotation in radians

        Returns:
            np.ndarray: a 3D homogeous transform matrix
        """
        return _scenepic.rotation_about_z(angle)

    look_at_rotation = _scenepic.look_at_rotation
    euler_angles_to_matrix = _scenepic.euler_angles_to_matrix
    quaternion_to_matrix = _scenepic.quaternion_to_matrix
    gl_projection = _scenepic.gl_projection
    gl_world_to_camera = _scenepic.gl_world_to_camera
    quaternion_multiply = _scenepic.quaternion_multiply


class LoopSubdivStencil:
    """Backwards compatibility wrapper for LoopSubdivisionStencil.

    Args:
        triangles (np.ndarray): an Nx3 numpy integer array specifying vertex
                                indices for each triangle.
        steps (int, optional): specifies how many steps of subdivision to apply.
                               Defaults to 1.
        project_to_limit (bool, optional): specifies whether the vertices should be
                                           projected onto the limit surface in the
                                           final step of subdivision. Defaults to False.
    """

    def __init__(self, triangles: np.ndarray, steps=1, project_to_limit=False):
        self._stencil = LoopSubdivisionStencil.create(triangles, steps, project_to_limit)

    def apply(self, vertices: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Apply a pre-computed subdivision stencil to a set of vertices.

        The same stencil can be reused across many sets of vertices,
        assuming the triangle topology does not change.

        Args:
            vertices (np.ndarray): float32 matrix of [N, 3] vertex positions

        Returns:
            Tuple[np.ndarray, np.ndarray]: new vertices, new triangles
        """
        return self._stencil.apply(vertices), self._stencil.triangles
